from datetime import datetime

from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.forms import BaseForm, BooleanField, CharField, ChoiceField, DateField, Form, IntegerField, ModelChoiceField, ModelForm
from django.forms.utils import ErrorDict
from django.utils import timezone

from NEMO.models import Account, Alert, Comment, Consumable, ConsumableWithdraw, Project, SafetyIssue, ScheduledOutage, Task, TaskCategory, User, UserPreferences
from NEMO.utilities import bootstrap_primary_color, format_datetime


class UserForm(ModelForm):
	class Meta:
		model = User
		fields = ['username', 'first_name', 'last_name', 'email', 'badge_number', 'access_expiration', 'type', 'domain', 'is_active', 'training_required', 'physical_access_levels', 'qualifications', 'projects']


class ProjectForm(ModelForm):
	class Meta:
		model = Project
		fields = ['name', 'application_identifier', 'account', 'active']


class AccountForm(ModelForm):
	class Meta:
		model = Account
		fields = ['name', 'active']


class TaskForm(ModelForm):
	problem_category = ModelChoiceField(queryset=TaskCategory.objects.filter(stage=TaskCategory.Stage.INITIAL_ASSESSMENT), required=False, label="Problem category")
	resolution_category = ModelChoiceField(queryset=TaskCategory.objects.filter(stage=TaskCategory.Stage.COMPLETION), required=False, label="Resolution category")
	action = ChoiceField(choices=[('create', 'create'), ('update', 'update'), ('resolve', 'resolve')], label="Action")
	description = CharField(required=False, label="Description")

	class Meta:
		model = Task
		fields = ['tool', 'urgency', 'estimated_resolution_time', 'force_shutdown', 'safety_hazard']

	def __init__(self, user, *args, **kwargs):
		super(TaskForm, self).__init__(*args, **kwargs)
		self.user = user
		self.fields['tool'].required = False
		self.fields['urgency'].required = False

	def clean_description(self):
		return self.cleaned_data['description'].strip()

	def clean(self):
		if any(self.errors):
			return
		super(TaskForm, self).clean()
		action = self.cleaned_data['action']
		if action == 'create':
			if not self.cleaned_data['description']:
				raise ValidationError('You must describe the problem.')
		if action == 'resolve':
			if self.instance.cancelled or self.instance.resolved:
				raise ValidationError("This task can't be resolved because it is marked as 'cancelled' or 'resolved' already.")

	def save(self, commit=True):
		instance = super(TaskForm, self).save(commit=False)
		action = self.cleaned_data['action']
		description = self.cleaned_data['description']
		instance.problem_category = self.cleaned_data['problem_category']
		now = timezone.now()
		if action == 'create':
			instance.problem_description = description
			instance.urgency = Task.Urgency.HIGH if self.cleaned_data['force_shutdown'] or self.cleaned_data['safety_hazard'] else Task.Urgency.NORMAL
			instance.creator = self.user
		if action == 'update':
			instance.last_updated = timezone.now()
			instance.last_updated_by = self.user
			instance.cancelled = False
			instance.resolved = False
			if description:
				preface = f'On {format_datetime(now)} {self.user.get_full_name()} updated this task:\n'
				if instance.progress_description is None:
					instance.progress_description = preface + description
				else:
					instance.progress_description += '\n\n' + preface + description
				instance.progress_description = instance.progress_description.strip()
		if action == 'resolve':
			instance.cancelled = False
			instance.resolved = True
			instance.resolution_time = now
			instance.resolver = self.user
			if 'resolution_category' in self.cleaned_data:
				instance.resolution_category = self.cleaned_data['resolution_category']
			if 'description' in self.cleaned_data:
				if instance.resolution_description:
					preface = f'On {format_datetime(now)} {self.user.get_full_name()} updated the resolution information:\n'
					instance.resolution_description = (instance.resolution_description + '\n\n' + preface + self.cleaned_data['description']).strip()
				else:
					instance.resolution_description = self.cleaned_data['description']
		return super(TaskForm, self).save(commit=True)


class CommentForm(ModelForm):
	class Meta:
		model = Comment
		fields = ['tool', 'content']

	expiration = IntegerField(label="Expiration date", min_value=0)


class SafetyIssueCreationForm(ModelForm):
	report_anonymously = BooleanField(required=False, initial=False)

	class Meta:
		model = SafetyIssue
		fields = ['reporter', 'concern', 'location']

	def __init__(self, user, *args, **kwargs):
		super(SafetyIssueCreationForm, self).__init__(*args, **kwargs)
		self.user = user

	def clean_update(self):
		return self.cleaned_data['concern'].strip()

	def clean_location(self):
		return self.cleaned_data['location'].strip()

	def save(self, commit=True):
		instance = super(SafetyIssueCreationForm, self).save(commit=False)
		if not self.cleaned_data['report_anonymously']:
			self.instance.reporter = self.user
		if commit:
			instance.save()
		return super(SafetyIssueCreationForm, self).save(commit=commit)


class SafetyIssueUpdateForm(ModelForm):
	update = CharField(required=False, label="Update")

	class Meta:
		model = SafetyIssue
		fields = ['resolved', 'visible']

	def __init__(self, user, *args, **kwargs):
		super(SafetyIssueUpdateForm, self).__init__(*args, **kwargs)
		self.user = user

	def clean_update(self):
		return self.cleaned_data['update'].strip()

	def save(self, commit=True):
		instance = super(SafetyIssueUpdateForm, self).save(commit=False)
		progress_type = 'resolved' if self.cleaned_data['resolved'] else 'updated'
		if progress_type == 'resolved':
			instance.resolution = self.cleaned_data['update']
			instance.resolution_time = timezone.now()
			instance.resolver = self.user
		if progress_type == 'updated' and self.cleaned_data['update']:
			progress = 'On ' + format_datetime(timezone.now()) + ' ' + self.user.get_full_name() + ' updated this issue:\n' + self.cleaned_data['update']
			if instance.progress:
				instance.progress += '\n\n' + progress
			else:
				instance.progress = progress
		return super(SafetyIssueUpdateForm, self).save(commit=commit)


class ConsumableWithdrawForm(ModelForm):
	class Meta:
		model = ConsumableWithdraw
		fields = ['customer', 'project', 'consumable', 'quantity']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.fields['consumable'].queryset = Consumable.objects.filter(visible=True)

	def clean_customer(self):
		customer = self.cleaned_data['customer']
		if not customer.is_active:
			raise ValidationError('A consumable withdraw was requested for an inactive user. Only active users may withdraw consumables.')
		return customer

	def clean_project(self):
		project = self.cleaned_data['project']
		if not project.active:
			raise ValidationError('A consumable may only be billed to an active project. The user\'s project is inactive.')
		if not project.account.active:
			raise ValidationError('A consumable may only be billed to a project that belongs to an active account. The user\'s account is inactive.')
		return project

	def clean_quantity(self):
		quantity = self.cleaned_data['quantity']
		if quantity < 1:
			raise ValidationError('Please specify a valid quantity of items to withdraw.')
		return quantity

	def clean(self):
		if any(self.errors):
			return
		super(ConsumableWithdrawForm, self).clean()
		quantity = self.cleaned_data['quantity']
		consumable = self.cleaned_data['consumable']
		if quantity > consumable.quantity:
			raise ValidationError('The withdraw was not processed because there are not enough "' + consumable.name + '". (There current quantity in stock is ' + str(consumable.quantity) + '). Please order more as soon as possible.')
		customer = self.cleaned_data['customer']
		project = self.cleaned_data['project']
		if project not in customer.active_projects():
			raise ValidationError('{} is not a member of the project {}. Users can only bill to projects they belong to.'.format(customer, project))


class ReservationAbuseForm(Form):
	cancellation_horizon = IntegerField(initial=6, min_value=1)
	cancellation_penalty = IntegerField(initial=10)
	target = IntegerField(required=False)
	start = DateField(initial=timezone.now().replace(day=1).date())
	end = DateField(initial=timezone.now().date())

	def clean_cancellation_horizon(self):
		# Convert the cancellation horizon from hours to seconds.
		return self.cleaned_data['cancellation_horizon'] * 60 * 60

	def clean_start(self):
		start = self.cleaned_data['start']
		return timezone.make_aware(datetime(year=start.year, month=start.month, day=start.day, hour=0, minute=0, second=0, microsecond=0), timezone.get_current_timezone())

	def clean_end(self):
		end = self.cleaned_data['end']
		return timezone.make_aware(datetime(year=end.year, month=end.month, day=end.day, hour=23, minute=59, second=59, microsecond=999999), timezone.get_current_timezone())


class EmailBroadcastForm(Form):
	subject = CharField(required=False)
	color = ChoiceField(choices=((bootstrap_primary_color('info'), 'info'), (bootstrap_primary_color('success'), 'success'), (bootstrap_primary_color('warning'), 'warning'), (bootstrap_primary_color('danger'), 'danger')))
	title = CharField(required=False)
	greeting = CharField(required=False)
	contents = CharField(required=False)
	copy_me = BooleanField(initial=True)

	audience = ChoiceField([('tool', 'tool'), ('project', 'project'), ('account', 'account')])
	selection = IntegerField()
	only_active_users = BooleanField(initial=True)

	def clean_title(self):
		return self.cleaned_data['title'].upper()


class AlertForm(ModelForm):
	class Meta:
		model = Alert
		fields = ['title', 'contents', 'debut_time', 'expiration_time']


class ScheduledOutageForm(ModelForm):
	def __init__(self, *positional_arguments, **keyword_arguments):
		super().__init__(*positional_arguments, **keyword_arguments)
		self.fields['details'].required = True

	class Meta:
		model = ScheduledOutage
		fields = ['details', 'start', 'end', 'resource', 'category']


class UserPreferencesForm(ModelForm):
	class Meta:
		model = UserPreferences
		fields = ['attach_created_reservation', 'attach_cancelled_reservation']


def nice_errors(form, non_field_msg='General form errors'):
	result = ErrorDict()
	if isinstance(form, BaseForm):
		for field, errors in form.errors.items():
			if field == NON_FIELD_ERRORS:
				key = non_field_msg
			else:
				key = form.fields[field].label
			result[key] = errors
	return result



import os
import sys

if __name__ == "__main__":
	os.environ.setdefault("DJANGO_SETTINGS_MODULE", "settings")
	from django.core.management import execute_from_command_line
	execute_from_command_line(sys.argv)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-08-15 15:23
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0002_version_1_1_0'),
    ]

    operations = [
        migrations.CreateModel(
            name='ScheduledOutageCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200)),
            ],
            options={
                'verbose_name_plural': 'Scheduled outage categories',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='TaskHistory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.CharField(help_text="A text description of the task's status", max_length=200)),
                ('time', models.DateTimeField(auto_now_add=True, help_text='The date and time when the task status was changed')),
            ],
            options={
                'verbose_name_plural': 'task histories',
                'ordering': ['time'],
                'get_latest_by': 'time',
            },
        ),
        migrations.CreateModel(
            name='TaskStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200, unique=True)),
                ('notify_primary_tool_owner', models.BooleanField(default=False, help_text='Notify the primary tool owner when a task transitions to this status')),
                ('notify_secondary_tool_owner', models.BooleanField(default=False, help_text='Notify the secondary tool owner when a task transitions to this status')),
                ('notify_tool_notification_email', models.BooleanField(default=False, help_text='Send an email to the tool notification email address when a task transitions to this status')),
                ('custom_notification_email_address', models.EmailField(blank=True, help_text="Notify a custom email address when a task transitions to this status. Leave this blank if you don't need it.", max_length=254)),
                ('notification_message', models.TextField(blank=True)),
            ],
            options={
                'verbose_name_plural': 'task statuses',
                'ordering': ['name'],
            },
        ),
        migrations.AlterModelOptions(
            name='resourcecategory',
            options={'ordering': ['name'], 'verbose_name_plural': 'resource categories'},
        ),
        migrations.RemoveField(
            model_name='task',
            name='first_responder',
        ),
        migrations.RemoveField(
            model_name='task',
            name='first_response_time',
        ),
        migrations.RemoveField(
            model_name='task',
            name='status',
        ),
        migrations.AddField(
            model_name='consumable',
            name='visible',
            field=models.BooleanField(default=True),
        ),
        migrations.AddField(
            model_name='scheduledoutage',
            name='category',
            field=models.CharField(blank=True, help_text='A categorical reason for why this outage is scheduled. Useful for trend analytics.', max_length=200),
        ),
        migrations.AddField(
            model_name='scheduledoutage',
            name='resource',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.Resource'),
        ),
        migrations.AddField(
            model_name='task',
            name='cancelled',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='task',
            name='resolved',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='tool',
            name='post_usage_questions',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='usageevent',
            name='run_data',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='resource',
            name='restriction_message',
            field=models.TextField(blank=True, help_text='The message that is displayed to users on the tool control page when this resource is unavailable.'),
        ),
        migrations.AlterField(
            model_name='scheduledoutage',
            name='details',
            field=models.TextField(blank=True, help_text='A detailed description of why there is a scheduled outage, and what users can expect during the outage'),
        ),
        migrations.AlterField(
            model_name='scheduledoutage',
            name='title',
            field=models.CharField(help_text='A brief description to quickly inform users about the outage', max_length=100),
        ),
        migrations.AlterField(
            model_name='scheduledoutage',
            name='tool',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.Tool'),
        ),
        migrations.AddField(
            model_name='taskhistory',
            name='task',
            field=models.ForeignKey(help_text='The task that this historical entry refers to', on_delete=django.db.models.deletion.CASCADE, related_name='history', to='NEMO.Task'),
        ),
        migrations.AddField(
            model_name='taskhistory',
            name='user',
            field=models.ForeignKey(help_text='The user that changed the task to this status', on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
    ]



from logging import getLogger

from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.shortcuts import get_object_or_404, redirect, render
from django.template import Context, Template
from django.utils import timezone
from django.views.decorators.http import require_GET, require_POST

from NEMO.forms import TaskForm, nice_errors
from NEMO.models import Interlock, Reservation, SafetyIssue, Task, TaskCategory, TaskHistory, TaskStatus, UsageEvent
from NEMO.utilities import bootstrap_primary_color, format_datetime, send_mail
from NEMO.views.customization import get_customization, get_media_file_contents
from NEMO.views.safety import send_safety_email_notification
from NEMO.views.tool_control import determine_tool_status

tasks_logger = getLogger("NEMO.Tasks")


@login_required
@require_POST
def create(request):
	"""
	This function handles feedback from users. This could be a problem report or shutdown notification.
	"""
	form = TaskForm(request.user, data=request.POST)
	if not form.is_valid():
		dictionary = {
			'title': 'Task creation failed',
			'heading': 'Something went wrong while reporting the problem',
			'content': nice_errors(form).as_ul(),
		}
		return render(request, 'acknowledgement.html', dictionary)
	task = form.save()

	if not settings.ALLOW_CONDITIONAL_URLS and task.force_shutdown:
		dictionary = {
			'title': 'Task creation failed',
			'heading': 'Something went wrong while reporting the problem',
			'content': "Tool control is only available on campus. When creating a task, you can't force a tool shutdown while using NEMO off campus.",
		}
		return render(request, 'acknowledgement.html', dictionary)

	if task.force_shutdown:
		# Shut down the tool.
		task.tool.operational = False
		task.tool.save()
		# End any usage events in progress for the tool.
		UsageEvent.objects.filter(tool=task.tool, end=None).update(end=timezone.now())
		# Lock the interlock for this tool.
		try:
			tool_interlock = Interlock.objects.get(tool__id=task.tool.id)
			tool_interlock.lock()
		except Interlock.DoesNotExist:
			pass

	if task.safety_hazard:
		concern = 'This safety issue was automatically created because a ' + str(task.tool).lower() + ' problem was identified as a safety hazard.\n\n'
		concern += task.problem_description
		issue = SafetyIssue.objects.create(reporter=request.user, location=task.tool.location, concern=concern)
		send_safety_email_notification(request, issue)

	send_new_task_emails(request, task)
	set_task_status(request, task, request.POST.get('status'), request.user)
	return redirect('tool_control')


def send_new_task_emails(request, task):
	message = get_media_file_contents('new_task_email.html')
	if message:
		dictionary = {
			'template_color': bootstrap_primary_color('danger') if task.force_shutdown else bootstrap_primary_color('warning'),
			'user': request.user,
			'task': task,
			'tool': task.tool,
			'tool_control_absolute_url': request.build_absolute_uri(task.tool.get_absolute_url())
		}
		# Send an email to the appropriate NanoFab staff that a new task has been created:
		subject = ('SAFETY HAZARD: ' if task.safety_hazard else '') + task.tool.name + (' shutdown' if task.force_shutdown else ' problem')
		message = Template(message).render(Context(dictionary))
		recipients = tuple([r for r in [task.tool.primary_owner.email, *task.tool.backup_owners.all().values_list('email', flat=True), task.tool.notification_email_address] if r])
		send_mail(subject, message, request.user.email, recipients)

	# Send an email to any user (excluding staff) with a future reservation on the tool:
	user_office_email = get_customization('user_office_email_address')
	message = get_media_file_contents('new_task_email.html')
	if user_office_email and message:
		upcoming_reservations = Reservation.objects.filter(start__gt=timezone.now(), cancelled=False, tool=task.tool, user__is_staff=False)
		for reservation in upcoming_reservations:
			if not task.tool.operational:
				subject = reservation.tool.name + " reservation problem"
				rendered_message = Template(message).render(Context({'reservation': reservation, 'template_color': bootstrap_primary_color('danger'), 'fatal_error': True}))
			else:
				subject = reservation.tool.name + " reservation warning"
				rendered_message = Template(message).render(Context({'reservation': reservation, 'template_color': bootstrap_primary_color('warning'), 'fatal_error': False}))
			reservation.user.email_user(subject, rendered_message, user_office_email)


@login_required
@require_POST
def cancel(request, task_id):
	task = get_object_or_404(Task, id=task_id)
	if task.cancelled or task.resolved:
		dictionary = {
			'title': 'Task cancellation failed',
			'heading': 'You cannot cancel this task',
			'content': 'The status of this task has been changed so you may no longer cancel it.',
		}
		return render(request, 'acknowledgement.html', dictionary)
	if task.creator.id != request.user.id:
		dictionary = {
			'title': 'Task cancellation failed',
			'heading': 'You cannot cancel this task',
			'content': 'You may only cancel a tasks that you created.',
		}
		return render(request, 'acknowledgement.html', dictionary)
	task.cancelled = True
	task.resolver = request.user
	task.resolution_time = timezone.now()
	task.save()
	determine_tool_status(task.tool)
	return redirect('tool_control')


def send_task_updated_email(task, url):
	if not hasattr(settings, 'LAB_MANAGERS'):
		return

	task.refresh_from_db()
	if task.resolved:
		task_user = task.resolver
		task_status = 'resolved'
	else:
		task_user = task.last_updated_by
		task_status = 'updated'
	message = f"""
A task for the {task.tool} was just modified by {task_user}.

The latest update is at the bottom of the description. The entirety of the task status follows: 

Task problem description:
{task.problem_description}

Task progress description:
{task.progress_description}

Task resolution description:
{task.resolution_description}

Visit {url} to view the tool control page for the task.
"""
	send_mail(f'{task.tool} task {task_status}', message, settings.SERVER_EMAIL, settings.LAB_MANAGERS)


@staff_member_required(login_url=None)
@require_POST
def update(request, task_id):
	task = get_object_or_404(Task, id=task_id)
	form = TaskForm(request.user, data=request.POST, instance=task)
	next_page = request.POST.get('next_page', 'tool_control')
	if not form.is_valid():
		dictionary = {
			'title': 'Task update failed',
			'heading': 'Invalid task form data',
			'content': str(form.errors),
		}
		return render(request, 'acknowledgement.html', dictionary)
	form.save()
	set_task_status(request, task, request.POST.get('status'), request.user)
	determine_tool_status(task.tool)
	try:
		send_task_updated_email(task, request.build_absolute_uri(task.tool.get_absolute_url()))
	except Exception as error:
		error_message = 'NEMO was unable to send the task updated email. The error message that NEMO received is: ' + str(error)
		tasks_logger.exception(error_message)
		pass
	if next_page == 'maintenance':
		return redirect('maintenance')
	else:
		return redirect('tool_control')


@staff_member_required(login_url=None)
@require_GET
def task_update_form(request, task_id):
	task = get_object_or_404(Task, id=task_id)
	categories = TaskCategory.objects.filter(stage=TaskCategory.Stage.INITIAL_ASSESSMENT)
	dictionary = {
		'categories': categories,
		'urgency': Task.Urgency.Choices,
		'task': task,
		'task_statuses': TaskStatus.objects.all(),
	}
	return render(request, 'tasks/update.html', dictionary)


@staff_member_required(login_url=None)
@require_GET
def task_resolution_form(request, task_id):
	task = get_object_or_404(Task, id=task_id)
	categories = TaskCategory.objects.filter(stage=TaskCategory.Stage.COMPLETION)
	dictionary = {
		'categories': categories,
		'task': task,
	}
	return render(request, 'tasks/resolve.html', dictionary)


def set_task_status(request, task, status_name, user):
	if not status_name:
		return

	if not user.is_staff:
		raise ValueError("Only staff can set task status")

	status = TaskStatus.objects.get(name=status_name)
	TaskHistory.objects.create(task=task, status=status_name, user=user)

	status_message = f'On {format_datetime(timezone.now())}, {user.get_full_name()} set the status of this task to "{status_name}".'
	task.progress_description = status_message if task.progress_description is None else task.progress_description + '\n\n' + status_message
	task.save()

	message = get_media_file_contents('task_status_notification.html')
	if not message:
		return

	dictionary = {
		'template_color': bootstrap_primary_color('success'),
		'title': f'{task.tool} task notification',
		'status_message': status_message,
		'notification_message': status.notification_message,
		'task': task,
		'tool_control_absolute_url': request.build_absolute_uri(task.tool.get_absolute_url())
	}
	# Send an email to the appropriate NanoFab staff that a new task has been created:
	subject = f'{task.tool} task notification'
	message = Template(message).render(Context(dictionary))
	recipients = [
		task.tool.primary_tool_owner.email if status.notify_primary_tool_owner else None,
		task.tool.notification_email_address if status.notify_tool_notification_email else None,
		status.custom_notification_email_address
	]
	if status.notify_backup_tool_owners:
		recipients += task.tool.backup_tool_owners.values_list('email')
	recipients = filter(None, recipients)
	send_mail(subject, message, user.email, recipients)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2019-09-10 12:46
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0007_version_1_15_0'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserPreferences',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('attach_created_reservation', models.BooleanField(default=False, help_text='Whether or not to send a calendar invitation when creating a new reservation', verbose_name='created_reservation_invite')),
                ('attach_cancelled_reservation', models.BooleanField(default=False, help_text='Whether or not to send a calendar invitation when cancelling a reservation', verbose_name='cancelled_reservation_invite')),
            ],
            options={
                'verbose_name': 'User preferences',
                'verbose_name_plural': 'User preferences',
            },
        ),
        migrations.AlterField(
            model_name='tool',
            name='policy_off_between_times',
            field=models.BooleanField(default=False, help_text='Check this box to disable policy rules every day between the given times'),
        ),
        migrations.AlterField(
            model_name='tool',
            name='policy_off_weekend',
            field=models.BooleanField(default=False, help_text='Whether or not policy rules should be enforced on weekends'),
        ),
        migrations.AddField(
            model_name='user',
            name='preferences',
            field=models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.UserPreferences'),
        ),
    ]



from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.http import JsonResponse
from django.shortcuts import render, get_object_or_404
from django.views.decorators.http import require_GET

from NEMO.models import User


@staff_member_required(login_url=None)
@require_GET
def get_projects(request):
	""" Gets a list of all active projects for a specific user. This is only accessible by staff members. """
	user = get_object_or_404(User, id=request.GET.get('user_id', None))
	projects = user.active_projects()
	source_template = request.GET.get('source_template')
	if source_template == 'training':
		entry_number = int(request.GET['entry_number'])
		return render(request, 'training/get_projects.html', {'projects': projects, 'entry_number': entry_number})
	elif source_template == 'staff_charges':
		return render(request, 'staff_charges/get_projects.html', {'projects': projects})
	return JsonResponse(dict(projects=list(projects.values('id', 'name'))))


@staff_member_required(login_url=None)
@require_GET
def get_projects_for_tool_control(request):
	user_id = request.GET.get('user_id')
	user = get_object_or_404(User, id=user_id)
	return render(request, 'tool_control/get_projects.html', {'active_projects': user.active_projects(), 'user_id': user_id})


@login_required
@require_GET
def get_projects_for_self(request):
	""" Gets a list of all active projects for the current user. """
	return render(request, 'tool_control/get_projects.html', {'active_projects': request.user.active_projects(), 'user_id': request.user.id})



from django.conf import settings
from django.conf.urls import include, url
from django.contrib import admin
from django.contrib.auth.decorators import login_required
from django.views.static import serve
from rest_framework import routers

from NEMO.views import abuse, accounts_and_projects, alerts, api, area_access, authentication, calendar, configuration_agenda, consumables, contact_staff, customization, email, feedback, get_projects, history, jumbotron, kiosk, landing, maintenance, mobile, usage, news, qualifications, remote_work, resources, safety, sidebar, staff_charges, status_dashboard, tasks, tool_control, training, tutorials, users

# Use our custom login page instead of Django's built-in one.
admin.site.login = login_required(admin.site.login)

# REST API URLs
router = routers.DefaultRouter()
router.register(r'users', api.UserViewSet)
router.register(r'projects', api.ProjectViewSet)
router.register(r'accounts', api.AccountViewSet)
router.register(r'tools', api.ToolViewSet)
router.register(r'reservations', api.ReservationViewSet)
router.register(r'usage_events', api.UsageEventViewSet)
router.register(r'area_access_records', api.AreaAccessRecordViewSet)
router.register(r'tasks', api.TaskViewSet)
router.register(r'scheduled_outages', api.ScheduledOutageViewSet)

urlpatterns = [
	# Authentication & error pages:
	url(r'^login/$', authentication.login_user, name='login'),
	url(r'^logout/$', authentication.logout_user, name='logout'),

	# Root URL defaults to the calendar page on desktop systems, and the mobile homepage for mobile devices:
	url(r'^$', landing.landing, name='landing'),

	# Get a list of projects for a user:
	url(r'^get_projects/$', get_projects.get_projects, name='get_projects'),
	url(r'^get_projects_for_tool_control/$', get_projects.get_projects_for_tool_control, name='get_projects_for_tool_control'),
	url(r'^get_projects_for_self/$', get_projects.get_projects_for_self, name='get_projects_for_self'),

	# Tool control:
	url(r'^tool_control/(?P<tool_id>\d+)/$', tool_control.tool_control, name='tool_control'),
	url(r'^tool_control/$', tool_control.tool_control, name='tool_control'),
	url(r'^tool_status/(?P<tool_id>\d+)/$', tool_control.tool_status, name='tool_status'),
	url(r'^use_tool_for_other/$', tool_control.use_tool_for_other, name='use_tool_for_other'),
	url(r'^tool_configuration/$', tool_control.tool_configuration, name='tool_configuration'),
	url(r'^create_comment/$', tool_control.create_comment, name='create_comment'),
	url(r'^hide_comment/(?P<comment_id>\d+)/$', tool_control.hide_comment, name='hide_comment'),
	url(r'^enable_tool/(?P<tool_id>\d+)/user/(?P<user_id>\d+)/project/(?P<project_id>\d+)/staff_charge/(?P<staff_charge>(true|false))/$', tool_control.enable_tool, name='enable_tool'),
	url(r'^disable_tool/(?P<tool_id>\d+)/$', tool_control.disable_tool, name='disable_tool'),
	url(r'^past_comments_and_tasks/$', tool_control.past_comments_and_tasks, name='past_comments_and_tasks'),
	url(r'^ten_most_recent_past_comments_and_tasks/(?P<tool_id>\d+)/$', tool_control.ten_most_recent_past_comments_and_tasks, name='ten_most_recent_past_comments_and_tasks'),

	# Tasks:
	url(r'^create_task/$', tasks.create, name='create_task'),
	url(r'^cancel_task/(?P<task_id>\d+)/$', tasks.cancel, name='cancel_task'),
	url(r'^update_task/(?P<task_id>\d+)/$', tasks.update, name='update_task'),
	url(r'^task_update_form/(?P<task_id>\d+)/$', tasks.task_update_form, name='task_update_form'),
	url(r'^task_resolution_form/(?P<task_id>\d+)/$', tasks.task_resolution_form, name='task_resolution_form'),

	# Calendar:
	url(r'^calendar/(?P<tool_id>\d+)/$', calendar.calendar, name='calendar'),
	url(r'^calendar/$', calendar.calendar, name='calendar'),
	url(r'^event_feed/$', calendar.event_feed, name='event_feed'),
	url(r'^create_reservation/$', calendar.create_reservation, name='create_reservation'),
	url(r'^create_outage/$', calendar.create_outage, name='create_outage'),
	url(r'^resize_reservation/$', calendar.resize_reservation, name='resize_reservation'),
	url(r'^resize_outage/$', calendar.resize_outage, name='resize_outage'),
	url(r'^move_reservation/$', calendar.move_reservation, name='move_reservation'),
	url(r'^move_outage/$', calendar.move_outage, name='move_outage'),
	url(r'^cancel_reservation/(?P<reservation_id>\d+)/$', calendar.cancel_reservation, name='cancel_reservation'),
	url(r'^cancel_outage/(?P<outage_id>\d+)/$', calendar.cancel_outage, name='cancel_outage'),
	url(r'^set_reservation_title/(?P<reservation_id>\d+)/$', calendar.set_reservation_title, name='set_reservation_title'),
	url(r'^event_details/reservation/(?P<reservation_id>\d+)/$', calendar.reservation_details, name='reservation_details'),
	url(r'^event_details/outage/(?P<outage_id>\d+)/$', calendar.outage_details, name='outage_details'),
	url(r'^event_details/usage/(?P<event_id>\d+)/$', calendar.usage_details, name='usage_details'),
	url(r'^event_details/area_access/(?P<event_id>\d+)/$', calendar.area_access_details, name='area_access_details'),
	url(r'^proxy_reservation/$', calendar.proxy_reservation, name='proxy_reservation'),

	# Qualifications:
	url(r'^qualifications/$', qualifications.qualifications, name='qualifications'),
	url(r'^modify_qualifications/$', qualifications.modify_qualifications, name='modify_qualifications'),
	url(r'^get_qualified_users/$', qualifications.get_qualified_users, name='get_qualified_users'),

	# Staff charges:
	url(r'^staff_charges/$', staff_charges.staff_charges, name='staff_charges'),
	url(r'^begin_staff_charge/$', staff_charges.begin_staff_charge, name='begin_staff_charge'),
	url(r'^end_staff_charge/$', staff_charges.end_staff_charge, name='end_staff_charge'),
	url(r'^begin_staff_area_charge/$', staff_charges.begin_staff_area_charge, name='begin_staff_area_charge'),
	url(r'^end_staff_area_charge/$', staff_charges.end_staff_area_charge, name='end_staff_area_charge'),

	# Status dashboard:
	url(r'^status_dashboard/$', status_dashboard.status_dashboard, name='status_dashboard'),
	url(r'^status_dashboard/(?P<tab>tools|occupancy)/$', status_dashboard.status_dashboard, name='status_dashboard_tab'),

	# Jumbotron:
	url(r'^jumbotron/$', jumbotron.jumbotron, name='jumbotron'),
	url(r'^jumbotron_content/$', jumbotron.jumbotron_content, name='jumbotron_content'),

	# Utility functions:
	url(r'^refresh_sidebar_icons/$', sidebar.refresh_sidebar_icons, name='refresh_sidebar_icons'),

	# NanoFab feedback
	url(r'^feedback/$', feedback.feedback, name='feedback'),

	# NanoFab rules tutorial
	# TODO: this should be removed, since this is really a job for a Learning Management System...
	url(r'^nanofab_rules_tutorial/$', tutorials.nanofab_rules, name='nanofab_rules'),

	# Configuration agenda for staff:
	url(r'^configuration_agenda/$', configuration_agenda.configuration_agenda, name='configuration_agenda'),
	url(r'^configuration_agenda/near_future/$', configuration_agenda.configuration_agenda, {'time_period': 'near_future'}, name='configuration_agenda_near_future'),

	# Email broadcasts:
	url(r'^get_email_form/$', email.get_email_form, name='get_email_form'),
	url(r'^get_email_form_for_user/(?P<user_id>\d+)/$', email.get_email_form_for_user, name='get_email_form_for_user'),
	url(r'^send_email/$', email.send_email, name='send_email'),
	url(r'^email_broadcast/$', email.email_broadcast, name='email_broadcast'),
	url(r'^email_broadcast/(?P<audience>tool|account|project)/$', email.email_broadcast, name='email_broadcast'),
	url(r'^compose_email/$', email.compose_email, name='compose_email'),
	url(r'^send_broadcast_email/$', email.send_broadcast_email, name='send_broadcast_email'),

	# Maintenance:
	url(r'^maintenance/(?P<sort_by>urgency|force_shutdown|tool|problem_category|last_updated|creation_time)/$', maintenance.maintenance, name='maintenance'),
	url(r'^maintenance/$', maintenance.maintenance, name='maintenance'),
	url(r'^task_details/(?P<task_id>\d+)/$', maintenance.task_details, name='task_details'),

	# Resources:
	url(r'^resources/$', resources.resources, name='resources'),
	url(r'^resources/modify/(?P<resource_id>\d+)/$', resources.modify_resource, name='modify_resource'),
	url(r'^resources/schedule_outage/$', resources.schedule_outage, name='schedule_resource_outage'),
	url(r'^resources/delete_scheduled_outage/(?P<outage_id>\d+)/$', resources.delete_scheduled_outage, name='delete_scheduled_resource_outage'),

	# Consumables:
	url(r'^consumables/$', consumables.consumables, name='consumables'),

	# Training:
	url(r'^training/$', training.training, name='training'),
	url(r'^training_entry/$', training.training_entry, name='training_entry'),
	url(r'^charge_training/$', training.charge_training, name='charge_training'),

	# Safety:
	url(r'^safety/$', safety.safety, name='safety'),
	url(r'^safety/resolved$', safety.resolved_safety_issues, name='resolved_safety_issues'),
	url(r'^safety/update/(?P<ticket_id>\d+)/$', safety.update_safety_issue, name='update_safety_issue'),

	# Mobile:
	url(r'^choose_tool/then/(?P<next_page>view_calendar|tool_control)/$', mobile.choose_tool, name='choose_tool'),
	url(r'^new_reservation/(?P<tool_id>\d+)/$', mobile.new_reservation, name='new_reservation'),
	url(r'^new_reservation/(?P<tool_id>\d+)/(?P<date>20\d\d-(0[1-9]|1[012])-(0[1-9]|[12][0-9]|3[01]))/$', mobile.new_reservation, name='new_reservation'),
	url(r'^make_reservation/$', mobile.make_reservation, name='make_reservation'),
	url(r'^view_calendar/(?P<tool_id>\d+)/$', mobile.view_calendar, name='view_calendar'),
	url(r'^view_calendar/(?P<tool_id>\d+)/(?P<date>20\d\d-(0[1-9]|1[012])-(0[1-9]|[12][0-9]|3[01]))/$', mobile.view_calendar, name='view_calendar'),

	# Contact staff:
	url(r'^contact_staff/$', contact_staff.contact_staff, name='contact_staff'),

	# Area access:
	url(r'^change_project/$', area_access.change_project, name='change_project'),
	url(r'^change_project/(?P<new_project>\d+)/$', area_access.change_project, name='change_project'),
	url(r'^force_area_logout/(?P<user_id>\d+)/$', area_access.force_area_logout, name='force_area_logout'),
	url(r'^self_log_in/$', area_access.self_log_in, name='self_log_in'),

	# NanoFab usage:
	url(r'^usage/$', usage.usage, name='usage'),
	url(r'^billing/$', usage.billing, name='billing'),

	# Alerts:
	url(r'^alerts/$', alerts.alerts, name='alerts'),
	url(r'^delete_alert/(?P<alert_id>\d+)/$', alerts.delete_alert, name='delete_alert'),

	# News:
	url(r'^news/$', news.view_recent_news, name='view_recent_news'),
	url(r'^news/archive/$', news.view_archived_news, name='view_archived_news'),
	url(r'^news/archive/(?P<page>\d+)/$', news.view_archived_news, name='view_archived_news'),
	url(r'^news/archive_story/(?P<story_id>\d+)/$', news.archive_story, name='archive_story'),
	url(r'^news/new/$', news.new_news_form, name='new_news_form'),
	url(r'^news/update/(?P<story_id>\d+)/$', news.news_update_form, name='news_update_form'),
	url(r'^news/publish/$', news.publish, name='publish_new_news'),
	url(r'^news/publish/(?P<story_id>\d+)/$', news.publish, name='publish_news_update'),

	# Media
	url(r'^media/(?P<path>.*)$', serve, {'document_root': settings.MEDIA_ROOT}, name='media'),

	# User Preferences
	url(r'^user_preferences/$', users.user_preferences, name='user_preferences')
]

if settings.ALLOW_CONDITIONAL_URLS:
	urlpatterns += [
		url(r'^admin/', include(admin.site.urls)),
		url(r'^api/', include(router.urls)),

		# Tablet area access
		url(r'^welcome_screen/(?P<door_id>\d+)/$', area_access.welcome_screen, name='welcome_screen'),
		url(r'^farewell_screen/(?P<door_id>\d+)/$', area_access.farewell_screen, name='farewell_screen'),
		url(r'^login_to_area/(?P<door_id>\d+)/$', area_access.login_to_area, name='login_to_area'),
		url(r'^logout_of_area/(?P<door_id>\d+)/$', area_access.logout_of_area, name='logout_of_area'),
		url(r'^open_door/(?P<door_id>\d+)/$', area_access.open_door, name='open_door'),

		# Tablet kiosk
		url(r'^kiosk/enable_tool/$', kiosk.enable_tool, name='enable_tool_from_kiosk'),
		url(r'^kiosk/disable_tool/$', kiosk.disable_tool, name='disable_tool_from_kiosk'),
		url(r'^kiosk/choices/$', kiosk.choices, name='kiosk_choices'),
		url(r'^kiosk/category_choices/(?P<category>.+)/(?P<user_id>\d+)/$', kiosk.category_choices, name='kiosk_category_choices'),
		url(r'^kiosk/tool_information/(?P<tool_id>\d+)/(?P<user_id>\d+)/(?P<back>back_to_start|back_to_category)/$', kiosk.tool_information, name='kiosk_tool_information'),
		url(r'^kiosk/(?P<location>.+)/$', kiosk.kiosk, name='kiosk'),
		url(r'^kiosk/$', kiosk.kiosk, name='kiosk'),

		# Area access
		url(r'^area_access/$', area_access.area_access, name='area_access'),
		url(r'^new_area_access_record/$', area_access.new_area_access_record, name='new_area_access_record'),

		# General area occupancy table, for use with Kiosk and Area Access tablets
		url(r'^occupancy/$', status_dashboard.occupancy, name='occupancy'),

		# Reminders and periodic events
		url(r'^email_reservation_reminders/$', calendar.email_reservation_reminders, name='email_reservation_reminders'),
		url(r'^email_usage_reminders/$', calendar.email_usage_reminders, name='email_usage_reminders'),
		url(r'^cancel_unused_reservations/$', calendar.cancel_unused_reservations, name='cancel_unused_reservations'),

		# Abuse:
		url(r'^abuse/$', abuse.abuse, name='abuse'),
		url(r'^abuse/user_drill_down/$', abuse.user_drill_down, name='user_drill_down'),

		# User management:
		url(r'^users/$', users.users, name='users'),
		url(r'^user/(?P<user_id>\d+|new)/', users.create_or_modify_user, name='create_or_modify_user'),
		url(r'^deactivate_user/(?P<user_id>\d+)/', users.deactivate, name='deactivate_user'),
		url(r'^reset_password/(?P<user_id>\d+)/$', users.reset_password, name='reset_password'),
		url(r'^unlock_account/(?P<user_id>\d+)/$', users.unlock_account, name='unlock_account'),

		# Account & project management:
		url(r'^accounts_and_projects/$', accounts_and_projects.accounts_and_projects, name='accounts_and_projects'),
		url(r'^project/(?P<identifier>\d+)/$', accounts_and_projects.accounts_and_projects, kwargs={'kind': 'project'}, name='project'),
		url(r'^account/(?P<identifier>\d+)/$', accounts_and_projects.accounts_and_projects, kwargs={'kind': 'account'}, name='account'),
		url(r'^toggle_active/(?P<kind>account|project)/(?P<identifier>\d+)/$', accounts_and_projects.toggle_active, name='toggle_active'),
		url(r'^create_project/$', accounts_and_projects.create_project, name='create_project'),
		url(r'^create_account/$', accounts_and_projects.create_account, name='create_account'),
		url(r'^remove_user/(?P<user_id>\d+)/from_project/(?P<project_id>\d+)/$', accounts_and_projects.remove_user_from_project, name='remove_user_from_project'),
		url(r'^add_user/(?P<user_id>\d+)/to_project/(?P<project_id>\d+)/$', accounts_and_projects.add_user_to_project, name='add_user_to_project'),

		# Account, project, and user history
		url(r'^history/(?P<item_type>account|project|user)/(?P<item_id>\d+)/$', history.history, name='history'),

		# Remote work:
		url(r'^remote_work/$', remote_work.remote_work, name='remote_work'),
		url(r'^validate_staff_charge/(?P<staff_charge_id>\d+)/$', remote_work.validate_staff_charge, name='validate_staff_charge'),
		url(r'^validate_usage_event/(?P<usage_event_id>\d+)/$', remote_work.validate_usage_event, name='validate_usage_event'),

		# Site customization:
		url(r'^customization/$', customization.customization, name='customization'),
		url(r'^customize/(?P<element>.+)/$', customization.customize, name='customize'),

		# Project Usage:
		url(r'^project_usage/$', usage.project_usage, name='project_usage'),
		url(r'^project_billing/$', usage.project_billing, name='project_billing'),
	]

if settings.DEBUG:
	# Static files
	url(r'^static/(?P<path>.*$)', serve, {'document_root': settings.STATIC_ROOT}, name='static'),

	try:
		# Django debug toolbar
		import debug_toolbar
		urlpatterns += [
			url(r'^__debug__/', include(debug_toolbar.urls)),
		]
	except ImportError:
		pass



from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.contenttypes.models import ContentType
from django.http import HttpResponseBadRequest
from django.shortcuts import get_object_or_404, render
from django.utils.text import capfirst
from django.views.decorators.http import require_GET

from NEMO.models import Account, Project, User, ActivityHistory, MembershipHistory


@staff_member_required(login_url=None)
@require_GET
def history(request, item_type, item_id):
	if item_type == "account":
		item = get_object_or_404(Account, id=item_id)
	elif item_type == "project":
		item = get_object_or_404(Project, id=item_id)
	elif item_type == "user":
		item = get_object_or_404(User, id=item_id)
	else:
		return HttpResponseBadRequest("Invalid history type")
	content_type = ContentType.objects.get_for_model(item)
	activity = ActivityHistory.objects.filter(object_id=item_id, content_type__id__exact=content_type.id)
	membership = MembershipHistory.objects.filter(parent_object_id=item_id, parent_content_type__id__exact=content_type.id)
	ownership = MembershipHistory.objects.filter(child_object_id=item_id, child_content_type__id__exact=content_type.id)
	# Iterate over all activity history and membership history for this object.
	action_list = []
	for a in activity:
		message = capfirst(content_type.name) + " "
		if a.action == ActivityHistory.Action.ACTIVATED:
			message += "activated."
		else:
			message += "deactivated."
		action_list.append({'date': a.date, 'authorizer': str(a.authorizer), 'message': message})
	for m in membership:
		message = capfirst(m.child_content_type.name) + " \"" + str(m.child_content_object) + "\" "
		if m.action:
			message += "added to"
		else:
			message += "removed from"
		message += " this " + content_type.name + "."
		action_list.append({'date': m.date, 'authorizer': str(m.authorizer), 'message': message})
	for o in ownership:
		message = "This " + content_type.name + " "
		if o.action:
			message += "now"
		else:
			message += "no longer"
		message += " belongs to " + o.parent_content_type.name + " \"" + o.parent_content_object.name + "\"."
		action_list.append({'date': o.date, 'authorizer': str(o.authorizer), 'message': message})
	# Sort the list of actions by date:
	action_list.sort(key=lambda x: x['date'])
	return render(request, 'history.html', {'action_list': action_list, 'name': str(item)})



from rest_framework.viewsets import ReadOnlyModelViewSet

from NEMO.filters import ReservationFilter, UsageEventFilter, AreaAccessRecordFilter, UserFilter
from NEMO.models import User, Project, Account, Reservation, UsageEvent, AreaAccessRecord, Task, ScheduledOutage, Tool
from NEMO.serializers import UserSerializer, ProjectSerializer, AccountSerializer, ReservationSerializer, UsageEventSerializer, AreaAccessRecordSerializer, TaskSerializer, ScheduledOutageSerializer, ToolSerializer


class UserViewSet(ReadOnlyModelViewSet):
	queryset = User.objects.all()
	serializer_class = UserSerializer
	filter_class = UserFilter


class ProjectViewSet(ReadOnlyModelViewSet):
	queryset = Project.objects.all()
	serializer_class = ProjectSerializer


class AccountViewSet(ReadOnlyModelViewSet):
	queryset = Account.objects.all()
	serializer_class = AccountSerializer


class ToolViewSet(ReadOnlyModelViewSet):
	queryset = Tool.objects.all()
	serializer_class = ToolSerializer


class ReservationViewSet(ReadOnlyModelViewSet):
	queryset = Reservation.objects.all()
	serializer_class = ReservationSerializer
	filter_class = ReservationFilter


class UsageEventViewSet(ReadOnlyModelViewSet):
	queryset = UsageEvent.objects.all()
	serializer_class = UsageEventSerializer
	filter_class = UsageEventFilter


class AreaAccessRecordViewSet(ReadOnlyModelViewSet):
	queryset = AreaAccessRecord.objects.all()
	serializer_class = AreaAccessRecordSerializer
	filter_class = AreaAccessRecordFilter


class TaskViewSet(ReadOnlyModelViewSet):
	queryset = Task.objects.all()
	serializer_class = TaskSerializer


class ScheduledOutageViewSet(ReadOnlyModelViewSet):
	queryset = ScheduledOutage.objects.all()
	serializer_class = ScheduledOutageSerializer



from logging import getLogger
from smtplib import SMTPException

from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.core.mail import EmailMultiAlternatives
from django.core.validators import validate_email
from django.http import HttpResponseBadRequest
from django.shortcuts import render, get_object_or_404
from django.template import Template, Context
from django.views.decorators.http import require_GET, require_POST

from NEMO.forms import EmailBroadcastForm
from NEMO.models import Tool, Account, Project, User
from NEMO.views.customization import get_media_file_contents


logger = getLogger(__name__)


@login_required
@require_GET
def get_email_form(request):
	recipient = request.GET.get('recipient', '')
	try:
		validate_email(recipient)
	except:
		return HttpResponseBadRequest('Recipient not valid.')
	return render(request, 'email/email_form.html', {'recipient': recipient})


@login_required
@require_GET
def get_email_form_for_user(request, user_id):
	recipient = get_object_or_404(User, id=user_id)
	return render(request, 'email/email_form.html', {'name': recipient.get_full_name(), 'recipient': recipient.email})


@login_required
@require_POST
def send_email(request):
	try:
		recipient = request.POST['recipient']
		validate_email(recipient)
		recipient_list = [recipient]
	except:
		return HttpResponseBadRequest('The intended recipient was not a valid email address. The email was not sent.')
	sender = request.user.email
	subject = request.POST.get('subject')
	body = request.POST.get('body')
	if request.POST.get('copy_me'):
		recipient_list.append(sender)
	try:
		email = EmailMultiAlternatives(subject, from_email=sender, bcc=recipient_list)
		email.attach_alternative(body, 'text/html')
		email.send()
	except SMTPException as error:
		error_message = 'NEMO was unable to send the email through the email server. The error message that NEMO received is: ' + str(error)
		logger.exception(error_message)
		dictionary = {
			'title': 'Email not sent',
			'heading': 'There was a problem sending your email',
			'content': error_message,
		}
		return render(request, 'acknowledgement.html', dictionary)
	dictionary = {
		'title': 'Email sent',
		'heading': 'Your email was sent',
	}
	return render(request, 'acknowledgement.html', dictionary)


@staff_member_required(login_url=None)
@require_GET
def email_broadcast(request, audience=''):
	dictionary = {}
	if audience == 'tool':
		dictionary['search_base'] = Tool.objects.filter(visible=True)
	elif audience == 'project':
		dictionary['search_base'] = Project.objects.filter(active=True, account__active=True)
	elif audience == 'account':
		dictionary['search_base'] = Account.objects.filter(active=True)
	dictionary['audience'] = audience
	return render(request, 'email/email_broadcast.html', dictionary)


@staff_member_required(login_url=None)
@require_GET
def compose_email(request):
	audience = request.GET.get('audience')
	selection = request.GET.get('selection')
	try:
		if audience == 'tool':
			users = User.objects.filter(qualifications__id=selection).distinct()
		elif audience == 'project':
			users = User.objects.filter(projects__id=selection).distinct()
		elif audience == 'account':
			users = User.objects.filter(projects__account__id=selection).distinct()
		else:
			dictionary = {'error': 'You specified an invalid audience'}
			return render(request, 'email/email_broadcast.html', dictionary)
	except:
		dictionary = {'error': 'You specified an invalid audience parameter'}
		return render(request, 'email/email_broadcast.html', dictionary)
	generic_email_sample = get_media_file_contents('generic_email.html')
	dictionary = {
		'audience': audience,
		'selection': selection,
		'users': users,
	}
	if generic_email_sample:
		generic_email_context = {
			'title': 'TITLE',
			'greeting': 'Greeting',
			'contents': 'Contents',
		}
		dictionary['generic_email_sample'] = Template(generic_email_sample).render(Context(generic_email_context))
	return render(request, 'email/compose_email.html', dictionary)


@staff_member_required(login_url=None)
@require_POST
def send_broadcast_email(request):
	if not get_media_file_contents('generic_email.html'):
		return HttpResponseBadRequest('Generic email template not defined. Visit the NEMO customizable_key_values page to upload a template.')
	form = EmailBroadcastForm(request.POST)
	if not form.is_valid():
		return render(request, 'email/compose_email.html', {'form': form})
	dictionary = {
		'title': form.cleaned_data['title'],
		'greeting': form.cleaned_data['greeting'],
		'contents': form.cleaned_data['contents'],
		'template_color': form.cleaned_data['color'],
	}
	content = get_media_file_contents('generic_email.html')
	content = Template(content).render(Context(dictionary))
	users = None
	audience = form.cleaned_data['audience']
	selection = form.cleaned_data['selection']
	active_choice = form.cleaned_data['only_active_users']
	try:
		if audience == 'tool':
			users = User.objects.filter(qualifications__id=selection)
		elif audience == 'project':
			users = User.objects.filter(projects__id=selection)
		elif audience == 'account':
			users = User.objects.filter(projects__account__id=selection)
		if active_choice:
			users = users.filter(is_active=True)
	except Exception as error:
		warning_message = 'Your email was not sent. There was a problem finding the users to send the email to.'
		dictionary = {'error': warning_message}
		logger.warning(warning_message + ' audience: {}, only_active: {}. The error message that NEMO received is: {}'.format(audience, active_choice, str(error)))
		return render(request, 'email/compose_email.html', dictionary)
	if not users:
		dictionary = {'error': 'The audience you specified is empty. You must send the email to at least one person.'}
		return render(request, 'email/compose_email.html', dictionary)
	subject = form.cleaned_data['subject']
	users = [x.email for x in users]
	if form.cleaned_data['copy_me']:
		users += [request.user.email]
	try:
		email = EmailMultiAlternatives(subject, from_email=request.user.email, bcc=set(users))
		email.attach_alternative(content, 'text/html')
		email.send()
	except SMTPException as error:
		error_message = 'NEMO was unable to send the email through the email server. The error message that NEMO received is: ' + str(error)
		logger.exception(error_message)
		dictionary = {
			'title': 'Email not sent',
			'heading': 'There was a problem sending your email',
			'content': error_message,
		}
		return render(request, 'acknowledgement.html', dictionary)
	dictionary = {
		'title': 'Email sent',
		'heading': 'Your email was sent',
	}
	return render(request, 'acknowledgement.html', dictionary)



from datetime import datetime, timedelta
from itertools import chain

from django.contrib.auth.decorators import login_required
from django.db.models import Q
from django.http import HttpResponseBadRequest
from django.shortcuts import render, get_object_or_404
from django.utils.dateparse import parse_time, parse_date
from django.views.decorators.http import require_GET, require_POST

from NEMO.models import Reservation, Tool, Project, ScheduledOutage
from NEMO.utilities import extract_date, localize, beginning_of_the_day, end_of_the_day
from NEMO.views.calendar import extract_configuration, determine_insufficient_notice
from NEMO.views.policy import check_policy_to_save_reservation


@login_required
@require_GET
def choose_tool(request, next_page):
	dictionary = {
		'tools': Tool.objects.filter(visible=True).order_by('category', 'name'),
	}
	if next_page == 'view_calendar':
		# If the user has no active projects then they're not allowed to make reservations. Redirect them home.
		if request.user.active_project_count() == 0:
			return render(request, 'mobile/no_active_projects.html')
		dictionary['title'] = "Which tool calendar would you like to view?"
		dictionary['next_page'] = 'view_calendar'
	elif next_page == 'tool_control':
		dictionary['title'] = "Which tool control page would you like to view?"
		dictionary['next_page'] = 'tool_control'
	return render(request, 'mobile/choose_tool.html', dictionary)


@login_required
@require_GET
def new_reservation(request, tool_id, date=None):
	# If the user has no active projects then they're not allowed to make reservations.
	if request.user.active_project_count() == 0:
		return render(request, 'mobile/no_active_projects.html')

	tool = get_object_or_404(Tool, id=tool_id)
	dictionary = tool.get_configuration_information(user=request.user, start=None)
	dictionary['tool'] = tool
	dictionary['date'] = date

	return render(request, 'mobile/new_reservation.html', dictionary)


@login_required
@require_POST
def make_reservation(request):
	""" Create a reservation for a user. """
	try:
		date = parse_date(request.POST['date'])
		start = localize(datetime.combine(date, parse_time(request.POST['start'])))
		end = localize(datetime.combine(date, parse_time(request.POST['end'])))
	except:
		return render(request, 'mobile/error.html', {'message': 'Please enter a valid date, start time, and end time for the reservation.'})
	tool = get_object_or_404(Tool, id=request.POST.get('tool_id'))
	# Create the new reservation:
	reservation = Reservation()
	reservation.user = request.user
	reservation.creator = request.user
	reservation.tool = tool
	reservation.start = start
	reservation.end = end
	reservation.short_notice = determine_insufficient_notice(tool, start)
	policy_problems, overridable = check_policy_to_save_reservation(None, reservation, request.user, False)

	# If there was a problem in saving the reservation then return the error...
	if policy_problems:
		return render(request, 'mobile/error.html', {'message': policy_problems[0]})

	# All policy checks have passed.
	try:
		reservation.project = Project.objects.get(id=request.POST['project_id'])
	except:
		if not request.user.is_staff:
			return render(request, 'mobile/error.html', {'message': 'You must specify a project for your reservation'})

	reservation.additional_information, reservation.self_configuration = extract_configuration(request)
	# Reservation can't be short notice if the user is configuring the tool themselves.
	if reservation.self_configuration:
		reservation.short_notice = False
	reservation.save_and_notify()
	return render(request, 'mobile/reservation_success.html', {'new_reservation': reservation})


@login_required
@require_GET
def view_calendar(request, tool_id, date=None):
	tool = get_object_or_404(Tool, id=tool_id)
	if date:
		try:
			date = extract_date(date)
		except:
			render(request, 'mobile/error.html', {'message': 'Invalid date requested for tool calendar'})
			return HttpResponseBadRequest()
	else:
		date = datetime.now()

	start = beginning_of_the_day(date, in_local_timezone=True)
	end = end_of_the_day(date, in_local_timezone=True)

	reservations = Reservation.objects.filter(tool=tool, cancelled=False, missed=False, shortened=False)
	# Exclude events for which the following is true:
	# The event starts and ends before the time-window, and...
	# The event starts and ends after the time-window.
	reservations = reservations.exclude(start__lt=start, end__lt=start)
	reservations = reservations.exclude(start__gt=end, end__gt=end)

	outages = ScheduledOutage.objects.filter(Q(tool=tool) | Q(resource__fully_dependent_tools__in=[tool]))
	outages = outages.exclude(start__lt=start, end__lt=start)
	outages = outages.exclude(start__gt=end, end__gt=end)

	events = list(chain(reservations, outages))
	events.sort(key=lambda x: x.start)

	dictionary = {
		'tool': tool,
		'previous_day': start - timedelta(days=1),
		'current_day': start,
		'current_day_string': date.strftime('%Y-%m-%d'),
		'next_day': start + timedelta(days=1),
		'events': events,
	}

	return render(request, 'mobile/view_calendar.html', dictionary)



import re

from django.http import HttpResponseForbidden
from django.contrib.auth.middleware import RemoteUserMiddleware
from django.conf import settings


class HTTPHeaderAuthenticationMiddleware(RemoteUserMiddleware):
	header = "HTTP_" + getattr(settings, "AUTHENTICATION_HEADER", "AUTHORIZATION")


class SessionTimeout:
	def __init__(self, get_response):
		self.get_response = get_response

	def __call__(self, request):
		return self.get_response(request)

	def process_view(self, request, view_function, view_ordered_arguments, view_named_arguments):
		"""
		This custom middleware exists to provide a smooth experience for public facing NEMO users.
		It processes all requests and redirects a user to the login page when their session times out.
		"""

		# If the user does not yet have a session then take no action.
		if not request.session:
			return None

		# If the user's session has timed out and this is an AJAX request then reply
		# with 403 forbidden because we don't want the AJAX request to produce the login page.
		# The base.html template has a global AJAX error callback to redirect to the logout
		# page when HTTP 403 is received.
		#
		# If the request is normal (instead of AJAX) and the user's session has expired
		# then the @login_required decorator will redirect them to the login page.
		if not request.user.is_authenticated:
			return HttpResponseForbidden() if request.is_ajax() else None

		# If the view is regularly polled by the webpage to update information then expiry refresh should be disabled.
		refresh_disabled = getattr(view_function, 'disable_session_expiry_refresh', False)
		if not refresh_disabled:
			request.session.modified = True


class DeviceDetectionMiddleware:
	def __init__(self, get_response):
		self.get_response = get_response
		self.mobile = re.compile('Mobile|Tablet|Android')

	def __call__(self, request):
		request.device = 'desktop'

		if 'HTTP_USER_AGENT' in request.META:
			user_agent = request.META['HTTP_USER_AGENT']
			if self.mobile.search(user_agent):
				request.device = 'mobile'

		return self.get_response(request)



from logging import getLogger

from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.shortcuts import render
from django.urls import reverse
from django.views.decorators.http import require_GET
from requests import get

from NEMO.models import AreaAccessRecord, ConsumableWithdraw, Reservation, StaffCharge, TrainingSession, UsageEvent, User, Project, Account
from NEMO.utilities import get_month_timeframe, month_list, parse_start_and_end_date


logger = getLogger(__name__)


# Class for Applications that can be used for autocomplete
class Application(object):
	def __init__(self, name):
		self.name = name
		self.id = name

	def __str__(self):
		return self.name


# We want to keep all the parameters of the request when switching tabs, so we are just replacing usage <-> billing urls
def get_url_for_other_tab(request):
	full_path_request = request.get_full_path()
	usage_url = reverse('usage')
	billing_url = reverse('billing')
	project_usage_url = reverse('project_usage')
	project_billing_url = reverse('project_billing')
	if project_usage_url in full_path_request:
		full_path_request = full_path_request.replace(project_usage_url, project_billing_url)
	elif project_billing_url in full_path_request:
		full_path_request = full_path_request.replace(project_billing_url, project_usage_url)
	elif usage_url in full_path_request:
		full_path_request = full_path_request.replace(usage_url, billing_url)
	elif billing_url in full_path_request:
		full_path_request = full_path_request.replace(billing_url, usage_url)
	return full_path_request


def get_project_applications():
	applications = []
	projects = Project.objects.filter(id__in=Project.objects.values('application_identifier').distinct().values_list('id', flat=True))
	for project in projects:
		if not any(list(filter(lambda app: app.name == project.application_identifier, applications))):
			applications.append(Application(project.application_identifier))
	return applications


def date_parameters_dictionary(request):
	if request.GET.get('start_date') and request.GET.get('end_date'):
		start_date, end_date = parse_start_and_end_date(request.GET.get('start_date'), request.GET.get('end_date'))
	else:
		start_date, end_date = get_month_timeframe()
	kind = request.GET.get("type")
	identifier = request.GET.get("id")
	dictionary = {
		'month_list': month_list(),
		'start_date': start_date,
		'end_date': end_date,
		'kind': kind,
		'identifier': identifier,
		'tab_url': get_url_for_other_tab(request)
	}
	return dictionary, start_date, end_date, kind, identifier


@login_required
@require_GET
def usage(request):
	base_dictionary, start_date, end_date, kind, identifier = date_parameters_dictionary(request)
	dictionary = {
		'area_access': AreaAccessRecord.objects.filter(customer=request.user, end__gt=start_date, end__lte=end_date),
		'consumables': ConsumableWithdraw.objects.filter(customer=request.user, date__gt=start_date, date__lte=end_date),
		'missed_reservations': Reservation.objects.filter(user=request.user, missed=True, end__gt=start_date, end__lte=end_date),
		'staff_charges': StaffCharge.objects.filter(customer=request.user, end__gt=start_date, end__lte=end_date),
		'training_sessions': TrainingSession.objects.filter(trainee=request.user, date__gt=start_date, date__lte=end_date),
		'usage_events': UsageEvent.objects.filter(user=request.user, end__gt=start_date, end__lte=end_date),
	}
	dictionary['no_charges'] = not (dictionary['area_access'] or dictionary['consumables'] or dictionary['missed_reservations'] or dictionary['staff_charges'] or dictionary['training_sessions'] or dictionary['usage_events'])
	return render(request, 'usage/usage.html', {**base_dictionary, **dictionary})


@login_required
@require_GET
def billing(request):
	base_dictionary, start_date, end_date, kind, identifier = date_parameters_dictionary(request)
	formatted_applications = ','.join(map(str, set(request.user.active_projects().values_list('application_identifier', flat=True))))
	try:
		billing_dictionary = billing_dict(start_date, end_date, request.user, formatted_applications)
		return render(request, 'usage/billing.html', {**base_dictionary, **billing_dictionary})
	except Exception as e:
		logger.warning(str(e))
		return render(request, 'usage/billing.html', base_dictionary)


@staff_member_required(login_url=None)
@require_GET
def project_usage(request):
	base_dictionary, start_date, end_date, kind, identifier = date_parameters_dictionary(request)

	projects = []
	selection = ''
	try:
		if kind == 'application':
			projects = Project.objects.filter(application_identifier=identifier)
			selection = identifier
		elif kind == 'project':
			projects = [Project.objects.get(id=identifier)]
			selection = projects[0].name
		elif kind == 'account':
			account = Account.objects.get(id=identifier)
			projects = Project.objects.filter(account=account)
			selection = account.name
	except:
		pass
	dictionary = {
		'accounts_and_applications': set(Account.objects.all()) | set(Project.objects.all()) | set(get_project_applications()),
		'area_access': AreaAccessRecord.objects.filter(project__in=projects, end__gt=start_date, end__lte=end_date) if projects else None,
		'consumables': ConsumableWithdraw.objects.filter(project__in=projects, date__gt=start_date, date__lte=end_date) if projects else None,
		'missed_reservations': Reservation.objects.filter(project__in=projects, missed=True, end__gt=start_date, end__lte=end_date) if projects else None,
		'staff_charges': StaffCharge.objects.filter(project__in=projects, end__gt=start_date, end__lte=end_date) if projects else None,
		'training_sessions': TrainingSession.objects.filter(project__in=projects, date__gt=start_date, date__lte=end_date) if projects else None,
		'usage_events': UsageEvent.objects.filter(project__in=projects, end__gt=start_date, end__lte=end_date) if projects else None,
		'project_autocomplete': True,
		'selection': selection
	}
	dictionary['no_charges'] = not (dictionary['area_access'] or dictionary['consumables'] or dictionary['missed_reservations'] or dictionary['staff_charges'] or dictionary['training_sessions'] or dictionary['usage_events'])
	return render(request, 'usage/usage.html', {**base_dictionary, **dictionary})


@staff_member_required(login_url=None)
@require_GET
def project_billing(request):
	base_dictionary, start_date, end_date, kind, identifier = date_parameters_dictionary(request)
	base_dictionary['project_autocomplete'] = True
	base_dictionary['accounts_and_applications'] = set(Account.objects.all()) | set(Project.objects.all()) | set(get_project_applications())

	project_id = None
	account_id = None
	formatted_applications = None
	selection = ''
	try:
		if kind == 'application':
			formatted_applications = identifier
			selection = identifier
		elif kind == 'project':
			projects = [Project.objects.get(id=identifier)]
			formatted_applications = projects[0].application_identifier
			project_id = identifier
			selection = projects[0].name
		elif kind == 'account':
			account = Account.objects.get(id=identifier)
			projects = Project.objects.filter(account=account, active=True, account__active=True)
			formatted_applications = ','.join(
				map(str, set(projects.values_list('application_identifier', flat=True)))) if projects else None
			account_id = account.id
			selection = account.name

		base_dictionary['selection'] = selection
		billing_dictionary = billing_dict(start_date, end_date, None, formatted_applications, project_id, account_id=account_id, force_pi=True)
		return render(request, 'usage/billing.html', {**base_dictionary, **billing_dictionary})
	except Exception as e:
		logger.warning(str(e))
		return render(request, 'usage/billing.html', base_dictionary)


def is_user_pi(user, application_pi_row):
	return application_pi_row is not None and (user.username == application_pi_row['username'] or (user.first_name == application_pi_row['first_name'] and user.last_name == application_pi_row['last_name']))


def billing_dict(start_date, end_date, user, formatted_applications, project_id=None, account_id=None, force_pi=None):
	dictionary = {}

	cost_activity_url = settings.BILLING_SERVICE['cost_activity_url']
	project_lead_url = settings.BILLING_SERVICE['project_lead_url']
	keyword_arguments = settings.BILLING_SERVICE['keyword_arguments']

	cost_activity_params = {
		'created_date_gte': f"'{start_date.strftime('%m/%d/%Y')}'",
		'created_date_lt': f"'{end_date.strftime('%m/%d/%Y')}'",
		'application_names': f"'{formatted_applications}'",
		'$format': 'json'
	}
	cost_activity_response = get(cost_activity_url, params=cost_activity_params, **keyword_arguments)
	cost_activity_data = cost_activity_response.json()['d']

	if force_pi is None:
		latest_pis_params = {'$format': 'json'}
		latest_pis_response = get(project_lead_url, params=latest_pis_params, **keyword_arguments)
		latest_pis_data = latest_pis_response.json()['d']

	project_totals = {}
	application_totals = {}
	account_totals = {}
	user_pi_applications = list()
	# Construct a tree of account, application, project, and member total spending
	cost_activities_tree = {}
	for activity in cost_activity_data:
		if (project_id and activity['project_id'] != str(project_id)) or (account_id and activity['account_id'] != str(account_id)):
			continue
		project_totals.setdefault(activity['project_id'], 0)
		application_totals.setdefault(activity['application_id'], 0)
		account_totals.setdefault(activity['account_id'], 0)
		account_key = (activity['account_id'], activity['account_name'])
		application_key = (activity['application_id'], activity['application_name'])
		project_key = (activity['project_id'], activity['project_name'])
		user_key = (activity['member_id'], User.objects.filter(id__in=[activity['member_id']]).first())
		user_is_pi = is_user_pi(user, next((x for x in latest_pis_data if x['application_name'] == activity['application_name']), None)) if force_pi is None else True
		if user_is_pi:
			user_pi_applications.append(activity['application_id'])
		if user_is_pi or str(user.id) == activity['member_id']:
			cost_activities_tree.setdefault((activity['account_id'], activity['account_name']), {})
			cost_activities_tree[account_key].setdefault(application_key, {})
			cost_activities_tree[account_key][application_key].setdefault(project_key, {})
			cost_activities_tree[account_key][application_key][project_key].setdefault(user_key, 0)
			cost = -activity['cost'] if activity['activity_type'] == 'refund_activity' else activity['cost']
			cost_activities_tree[account_key][application_key][project_key][user_key] = cost_activities_tree[account_key][application_key][project_key][user_key] + cost
			project_totals[activity['project_id']] = project_totals[activity['project_id']] + cost
			application_totals[activity['application_id']] = application_totals[activity['application_id']] + cost
			account_totals[activity['account_id']] = account_totals[activity['account_id']] + cost
	dictionary['spending'] = {
		'activities': cost_activities_tree,
		'project_totals': project_totals,
		'application_totals': application_totals,
		'account_totals': account_totals,
		'user_pi_applications': user_pi_applications
	} if cost_activities_tree else {'activities': {}}
	return dictionary



from json import dumps, loads

from django.utils import timezone
from django.utils.safestring import mark_safe

from NEMO.models import Consumable, ConsumableWithdraw
from NEMO.utilities import quiet_int


class DynamicForm:
	def __init__(self, questions):
		self.questions = loads(questions) if questions else None

	def render(self):
		if not self.questions:
			return ''

		result = ''
		for question in self.questions:
			if question['type'] == "radio":
				result += f'<div class="form-group">{question["title"]}'
				for choice in question['choices']:
					result += '<div class="radio">'
					required = 'required' if question['required'] else ''
					is_default_choice = 'checked' if question['default_choice'] == choice else ''
					result += f'<label><input type="radio" name="{question["name"]}" value="{choice}" {required} {is_default_choice}>{choice}</label>'
					result += '</div>'
				result += '</div>'
			elif question['type'] == "textbox":
				result += '<div class="form-group">'
				result += f'<label for="{question["name"]}">{question["title"]}</label>'
				input_group_required = True if 'prefix' in question or 'suffix' in question else False
				if input_group_required:
					result += f'<div class="input-group" style="max-width:{question["max-width"]}px">'
				if 'prefix' in question:
					result += f'<span class="input-group-addon">{question["prefix"]}</span>'
				required = 'required' if question['required'] is True else ''
				result += f'<input type="text" class="form-control" name="{question["name"]}" id="{question["name"]}" placeholder="{question["placeholder"]}" {required} style="max-width:{question["max-width"]}px" spellcheck="false" autocapitalize="off" autocomplete="off" autocorrect="off">'
				if 'suffix' in question:
					result += f'<span class="input-group-addon">{question["suffix"]}</span>'
				if input_group_required:
					result += '</div>'
				result += '</div>'

		return mark_safe(result)

	def extract(self, request):
		if not self.questions:
			return ''

		results = {}
		for question in self.questions:
			# Only record the answer when the question was answered. Discard questions that were left blank
			if request.POST.get(question['name']):
				results[question['name']] = request.POST[question['name']]
		return dumps(results, indent='\t', sort_keys=True) if len(results) else ''

	def charge_for_consumable(self, customer, merchant, project, run_data):
		try:
			run_data = loads(run_data)
		except:
			return
		for question in self.questions:
			if 'consumable' in question:
				try:
					consumable = Consumable.objects.get(name=question['consumable'])
					quantity = 0
					if question['type'] == 'textbox':
						if question['name'] in run_data:
							quantity = quiet_int(run_data[question['name']])
					elif question['type'] == 'radio':
						quantity = 1

					if quantity > 0:
						ConsumableWithdraw.objects.create(customer=customer, merchant=merchant, consumable=consumable, quantity=quantity, project=project, date=timezone.now())
				except:
					pass



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-10-03 17:07
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('NEMO', '0003_version_1_2_0'),
    ]

    operations = [
        migrations.CreateModel(
            name='News',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=200)),
                ('created', models.DateTimeField(help_text='The date and time this story was first published')),
                ('original_content', models.TextField(help_text="The content of the story when it was first published, useful for visually hiding updates 'in the middle' of the story")),
                ('all_content', models.TextField(help_text='The entire content of the story')),
                ('last_updated', models.DateTimeField(help_text='The date and time this story was last updated')),
                ('last_update_content', models.TextField(help_text="The most recent update to the story, useful for visually hiding updates 'in the middle' of the story")),
                ('archived', models.BooleanField(default=False, help_text="A story is removed from the 'Recent News' page when it is archived")),
                ('update_count', models.PositiveIntegerField(help_text="The number of times this story has been updated. When the number of updates is greater than 2, then only the original story and the latest update are displayed in the 'Recent News' page")),
            ],
            options={
                'verbose_name_plural': 'News',
                'ordering': ['-last_updated'],
            },
        ),
        migrations.CreateModel(
            name='Notification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('expiration', models.DateTimeField()),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='notifications', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.AddField(
            model_name='landingpagechoice',
            name='notifications',
            field=models.CharField(blank=True, choices=[('news', 'News creation and updates - notifies all users'), ('safetyissue', 'New safety issues - notifies staff only')], help_text="Displays a the number of new notifications for the user. For example, if the user has two unread news notifications then the number '2' would appear for the news icon on the landing page.", max_length=25, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.15 on 2018-11-21 14:36
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('NEMO', '0005_version_1_4_0'),
    ]

    operations = [
        migrations.AddField(
            model_name='tool',
            name='grant_badge_reader_access_upon_qualification',
            field=models.CharField(blank=True, help_text='Badge reader access is granted to the user upon qualification for this tool.', max_length=100, null=True),
        ),
        migrations.AddField(
            model_name='tool',
            name='grant_physical_access_level_upon_qualification',
            field=models.ForeignKey(blank=True, help_text='The designated physical access level is granted to the user upon qualification for this tool.', null=True, on_delete=django.db.models.deletion.CASCADE, to='NEMO.PhysicalAccessLevel'),
        ),
    ]



from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponseBadRequest
from django.shortcuts import render, get_object_or_404, redirect
from django.views.decorators.http import require_GET, require_POST, require_http_methods

from NEMO.forms import ProjectForm, AccountForm
from NEMO.models import Account, Project, User, MembershipHistory, ActivityHistory


@staff_member_required(login_url=None)
@require_GET
def accounts_and_projects(request, kind=None, identifier=None):
	try:
		if kind == 'project':
			account = Project.objects.get(id=identifier).account
		elif kind == 'account':
			account = Account.objects.get(id=identifier)
		else:
			account = None
	except:
		account = None
	dictionary = {
		'account': account,
		'account_list': Account.objects.all(),
		'accounts_and_projects': set(Account.objects.all()) | set(Project.objects.all()),
		'users': User.objects.all(),
	}
	return render(request, 'accounts_and_projects/accounts_and_projects.html', dictionary)


@staff_member_required(login_url=None)
@require_POST
def toggle_active(request, kind, identifier):
	if kind == 'account':
		entity = get_object_or_404(Account, id=identifier)
	elif kind == 'project':
		entity = get_object_or_404(Project, id=identifier)
	else:
		return HttpResponseBadRequest('Invalid entity for active toggle request.')
	entity.active = not entity.active
	entity.save()
	history = ActivityHistory()
	history.authorizer = request.user
	history.action = entity.active
	history.content_object = entity
	history.save()
	return redirect(request.META.get('HTTP_REFERER', 'accounts_and_projects'))


@staff_member_required(login_url=None)
@require_http_methods(['GET', 'POST'])
def create_project(request):
	dictionary = {
		'account_list': Account.objects.all(),
	}
	if request.method == 'GET':
		return render(request, 'accounts_and_projects/create_project.html', dictionary)
	form = ProjectForm(request.POST)
	if not form.is_valid():
		dictionary['form'] = form
		return render(request, 'accounts_and_projects/create_project.html', dictionary)
	project = form.save()
	account_history = MembershipHistory()
	account_history.authorizer = request.user
	account_history.action = MembershipHistory.Action.ADDED
	account_history.child_content_object = project
	account_history.parent_content_object = project.account
	account_history.save()
	project_history = ActivityHistory()
	project_history.authorizer = request.user
	project_history.action = project.active
	project_history.content_object = project
	project_history.save()
	return redirect('account', project.account.id)


@staff_member_required(login_url=None)
@require_http_methods(['GET', 'POST'])
def create_account(request):
	if request.method == 'GET':
		return render(request, 'accounts_and_projects/create_account.html')
	form = AccountForm(request.POST)
	if not form.is_valid():
		return render(request, 'accounts_and_projects/create_account.html', {'form': form})
	account = form.save()
	history = ActivityHistory()
	history.authorizer = request.user
	history.action = account.active
	history.content_object = account
	history.save()
	return redirect('account', account.id)


@staff_member_required(login_url=None)
@require_POST
def remove_user_from_project(request, user_id, project_id):
	user = get_object_or_404(User, id=user_id)
	project = get_object_or_404(Project, id=project_id)
	if project.user_set.filter(id=user.id).exists():
		history = MembershipHistory()
		history.action = MembershipHistory.Action.REMOVED
		history.authorizer = request.user
		history.parent_content_object = project
		history.child_content_object = user
		history.save()
		project.user_set.remove(user)
	dictionary = {
		'users': project.user_set.all(),
		'project': project
	}
	return render(request, 'accounts_and_projects/users_for_project.html', dictionary)


@staff_member_required(login_url=None)
@require_POST
def add_user_to_project(request, user_id, project_id):
	user = get_object_or_404(User, id=user_id)
	project = get_object_or_404(Project, id=project_id)
	if user not in project.user_set.all():
		history = MembershipHistory()
		history.action = MembershipHistory.Action.ADDED
		history.authorizer = request.user
		history.parent_content_object = project
		history.child_content_object = user
		history.save()
		project.user_set.add(user)
	dictionary = {
		'users': project.user_set.all(),
		'project': project
	}
	return render(request, 'accounts_and_projects/users_for_project.html', dictionary)



from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.http import Http404
from django.shortcuts import render, get_object_or_404, redirect
from django.utils import timezone
from django.views.decorators.http import require_POST, require_http_methods

from NEMO.forms import AlertForm
from NEMO.models import Alert


@staff_member_required(login_url=None)
@require_http_methods(['GET', 'POST'])
def alerts(request):
	alert_id = request.GET.get('alert_id') or request.POST.get('alert_id')
	try:
		alert = Alert.objects.get(id=alert_id)
	except:
		alert = None
	if request.method == 'GET':
		form = AlertForm(instance=alert)
	elif request.method == 'POST':
		form = AlertForm(data=request.POST, instance=alert)
		if form.is_valid():
			alert = form.save()
			if not alert.creator:
				alert.creator = request.user
			alert.save()
			form = AlertForm()
	else:
		form = AlertForm()
	dictionary = {
		'form': form,
		'editing': True if form.instance.id else False,
		'alerts': Alert.objects.filter(user=None)
	}
	delete_expired_alerts()
	return render(request, 'alerts.html', dictionary)


@login_required
@require_POST
def delete_alert(request, alert_id):
	try:
		alert = get_object_or_404(Alert, id=alert_id)
		if alert.user == request.user:  # Users can delete their own alerts
			alert.delete()
		elif alert.user is None and request.user.is_staff:  # Staff can delete global alerts
			alert.delete()
	except Http404:
		pass
	return redirect(request.META.get('HTTP_REFERER', 'landing'))


def delete_expired_alerts():
	Alert.objects.filter(expiration_time__lt=timezone.now()).delete()



from re import search
from urllib.parse import urljoin

import requests
from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.http import HttpResponseBadRequest
from django.shortcuts import render
from django.views.decorators.http import require_GET, require_POST

from NEMO.models import User, Tool, TrainingSession, Project, MembershipHistory


@staff_member_required(login_url=None)
@require_GET
def training(request):
	""" Present a web page to allow staff to charge training and qualify users on particular tools. """
	users = User.objects.filter(is_active=True).exclude(id=request.user.id)
	tools = Tool.objects.filter(visible=True)
	return render(request, 'training/training.html', {'users': users, 'tools': tools, 'charge_types': TrainingSession.Type.Choices})


@staff_member_required(login_url=None)
@require_GET
def training_entry(request):
	entry_number = int(request.GET['entry_number'])
	return render(request, 'training/training_entry.html', {'entry_number': entry_number, 'charge_types': TrainingSession.Type.Choices})


def is_valid_field(field):
	return search("^(chosen_user|chosen_tool|chosen_project|duration|charge_type|qualify)__[0-9]+$", field) is not None


@staff_member_required(login_url=None)
@require_POST
def charge_training(request):
	try:
		charges = {}
		for key, value in request.POST.items():
			if is_valid_field(key):
				attribute, separator, index = key.partition("__")
				index = int(index)
				if index not in charges:
					charges[index] = TrainingSession()
					charges[index].trainer = request.user
				if attribute == "chosen_user":
					charges[index].trainee = User.objects.get(id=value)
				if attribute == "chosen_tool":
					charges[index].tool = Tool.objects.get(id=value)
				if attribute == "chosen_project":
					charges[index].project = Project.objects.get(id=value)
				if attribute == "duration":
					charges[index].duration = int(value)
				if attribute == "charge_type":
					charges[index].type = int(value)
				if attribute == "qualify":
					charges[index].qualified = (value == "on")
		for c in charges.values():
			c.full_clean()
	except Exception:
		return HttpResponseBadRequest('An error occurred while processing the training charges. None of the charges were committed to the database. Please review the form for errors and omissions then submit the form again.')
	else:
		for c in charges.values():
			if c.qualified:
				qualify(c.trainer, c.trainee, c.tool)
			c.save()
		dictionary = {
			'title': 'Success!',
			'content': 'Training charges were successfully saved.',
			'redirect': '/',
		}
		return render(request, 'display_success_and_redirect.html', dictionary)


def qualify(authorizer, user, tool):
	if tool in user.qualifications.all():
		return
	user.qualifications.add(tool)
	entry = MembershipHistory()
	entry.authorizer = authorizer
	entry.parent_content_object = tool
	entry.child_content_object = user
	entry.action = entry.Action.ADDED
	entry.save()

	if tool.grant_physical_access_level_upon_qualification:
		if tool.grant_physical_access_level_upon_qualification not in user.physical_access_levels.all():
			user.physical_access_levels.add(tool.grant_physical_access_level_upon_qualification)
			entry = MembershipHistory()
			entry.authorizer = authorizer
			entry.parent_content_object = tool.grant_physical_access_level_upon_qualification
			entry.child_content_object = user
			entry.action = entry.Action.ADDED
			entry.save()

	if settings.IDENTITY_SERVICE['available']:
		if tool.grant_badge_reader_access_upon_qualification:
			parameters = {
				'username': user.username,
				'domain': user.domain,
				'requested_area': tool.grant_badge_reader_access_upon_qualification,
			}
			requests.put(urljoin(settings.IDENTITY_SERVICE['url'], '/add/'), data=parameters, timeout=3)
